from __future__ import annotations

import logging
from typing import cast

log = logging.getLogger(__name__)


from jsonschema import Draft202012Validator, ValidationError

from ._to_json import to_json_schema

__all__ = [
    "to_json_schema",
    "validate",
    "ConfigurationError",
    "MultiConfigurationError",
]


def validate(data: dict, schema: dict) -> None:
    """Validate the provided data against the given JSON schema.

    This function uses the Draft202012Validator to check if the data
    conforms to the specified schema. If there are any validation errors,
    it raises a ConfigurationError containing the details of the errors.

    Parameters
    ----------
    - data (dict): The data to be validated.
    - schema (dict): The JSON schema to validate against.

    Raises
    ------
    - ConfigurationError: If the data does not comply with the schema,
                          this error is raised with details of the violations.
    """
    schema = allow_none_in_schema(schema)
    validator = Draft202012Validator(schema)

    errors = list(validator.iter_errors(data))
    if errors:
        import json

        log.error("Validation errors in configuration data!")
        log.debug(f"Data: {data}")
        log.debug(f"Schema: {json.dumps(schema, indent=2)}")
        raise ConfigurationError.from_ValidationErrors(errors)


def allow_none_in_schema(schema: dict | list) -> dict:  # -> dict[Any, Any] | list[Any]:
    """Recursively modifies a JSON schema to allow `null` values for all fields."""
    if isinstance(schema, dict):
        # If current schema block has "type"
        if "type" in schema:
            # If the type is a list of types, add 'null' if it's not already present
            if isinstance(schema["type"], list):
                if "null" not in schema["type"]:
                    schema["type"].append("null")
            else:
                # Else, make it a list with the current type and 'null'
                schema["type"] = [schema["type"], "null"]

        # Recursively process properties in objects
        for key, value in schema.items():
            if isinstance(value, (dict, list)):
                allow_none_in_schema(value)

    elif isinstance(schema, list):
        for item in schema:
            allow_none_in_schema(item)

    return cast(dict, schema)


class ConfigurationError(Exception):
    """Exception raised for configuration validation errors."""

    def __init__(self, message="Configuration error", section=None):
        self.message = message
        self.section = section
        super().__init__(self.message)

    def __str__(self):
        """Return the error message."""
        return (
            f"{self.message} in section '{self.section}'"
            if self.section
            else self.message
        )

    @classmethod
    def from_ValidationErrors(cls, error: ValidationError | list[ValidationError]):
        """Create a ConfigurationError from a ValidationError or a list of ValidationErrors."""
        if isinstance(error, list):
            errors = []
            for e in error:
                path = [str(p) for p in e.path]
                errors.append(cls(e.message, ".".join(path)))
            return MultiConfigurationError(errors)
        else:
            return cls(error.message, str(error.path))


class MultiConfigurationError(Exception):
    """Exception raised for multiple configuration validation errors."""

    def __init__(self, errors: list[ConfigurationError]):
        self.errors = errors
        super().__init__("Multiple configuration errors")

    def __str__(self):
        """Return the error message."""
        return "\n".join([str(e) for e in self.errors])



from dataclasses import fields
from pprint import pprint
from typing import (
    Any,
    Dict,
    Literal,
    Optional,
    Sequence,
    TypedDict,
    Union,
    get_type_hints,
)
import pytest
from eyconf.validate import to_json_schema
from dataclasses import dataclass


class TestToSchema:
    """
    Test the dataclass to json schema function
    """

    @pytest.mark.parametrize(
        "as_dataclass",
        (True, False),
    )
    def test_primitives(self, as_dataclass):
        """Test a simple dataclass."""

        @dataclass
        class Primitives:
            foo: str
            bar: int
            baz: float
            qux: bool

        if not as_dataclass:
            Primitives = dataclass_to_typeddict(Primitives)  # type: ignore

        schema = to_json_schema(Primitives)

        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "string"},
                "bar": {"type": "integer"},
                "baz": {"type": "number"},
                "qux": {"type": "boolean"},
            },
            "required": ["foo", "bar", "baz", "qux"],
        }

    @pytest.mark.parametrize(
        "as_dataclass",
        (True, False),
    )
    def test_literal(self, as_dataclass):

        @dataclass
        class Schema:
            foo: Literal["bar", "baz"]

        if not as_dataclass:
            Schema = dataclass_to_typeddict(Schema)  # type: ignore

        schema = to_json_schema(Schema)

        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "string", "enum": ["bar", "baz"]},
            },
            "required": ["foo"],
        }

        @dataclass
        class InvalidLiteral:
            foo: Literal[1, "foo"]

        if not as_dataclass:
            InvalidLiteral = dataclass_to_typeddict(InvalidLiteral)  # type: ignore

        with pytest.raises(ValueError):
            to_json_schema(InvalidLiteral)

    @pytest.mark.parametrize(
        "as_dataclass",
        (True, False),
    )
    def test_optional(self, as_dataclass):

        @dataclass
        class Schema:
            foo: Optional[str]
            bar: Optional[int]
            baz: float

        if not as_dataclass:
            Schema = dataclass_to_typeddict(Schema)  # type: ignore

        schema = to_json_schema(Schema)
        print(schema)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "string"},
                "bar": {"type": "integer"},
                "baz": {"type": "number"},
            },
            "required": ["baz"],
        }

    @pytest.mark.parametrize(
        "as_dataclass",
        (True, False),
    )
    def test_union(self, as_dataclass):

        @dataclass
        class Schema:
            foo: str | int
            bar: int | float

        if not as_dataclass:
            Schema = dataclass_to_typeddict(Schema)  # type: ignore

        schema = to_json_schema(Schema)
        print(schema)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"anyOf": [{"type": "integer"}, {"type": "string"}]},
                "bar": {"anyOf": [{"type": "integer"}, {"type": "number"}]},
            },
            "required": ["foo", "bar"],
        }

    @pytest.mark.parametrize(
        "as_dataclass",
        (True, False),
    )
    def test_nested_dict(self, as_dataclass):

        @dataclass
        class Dict1:
            foo: str

        if not as_dataclass:
            Dict1 = dataclass_to_typeddict(Dict1)  # type: ignore

        @dataclass
        class NestedTyped:
            dict1: Dict1
            dict_opt: Optional[Dict1]
            dict_uni: Dict1 | str
            baz: Optional[float]

        if not as_dataclass:
            NestedTyped = dataclass_to_typeddict(NestedTyped)  # type: ignore

        schema = to_json_schema(NestedTyped)
        pprint(schema)
        assert schema["type"] == "object"
        assert schema["required"] == ["dict1", "dict_uni"]
        assert ["dict1", "dict_opt", "dict_uni", "baz"] == list(
            schema["properties"].keys()
        )

        dict1_obj = {
            "type": "object",
            "properties": {"foo": {"type": "string"}},
            "required": ["foo"],
        }

        assert schema["properties"]["dict1"] == dict1_obj
        assert schema["properties"]["dict_opt"] == dict1_obj

        assert "anyOf" in schema["properties"]["dict_uni"]
        assert (
            schema["properties"]["dict_uni"]["anyOf"][0] == dict1_obj
            or schema["properties"]["dict_uni"]["anyOf"][1] == dict1_obj
        )

    @pytest.mark.parametrize(
        "as_dataclass",
        (True, False),
    )
    def test_lists(self, as_dataclass):

        @dataclass
        class Schema:
            foo: list[str]
            bar: Optional[Sequence[int]]

        if not as_dataclass:
            Schema = dataclass_to_typeddict(Schema)  # type: ignore

        schema = to_json_schema(Schema)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "array", "items": {"type": "string"}},
                "bar": {"type": "array", "items": {"type": "integer"}},
            },
            "required": [
                "foo",
            ],
        }

    def test_not_required(self):
        from typing import TypedDict, NotRequired

        class MyTypedDict1(TypedDict):
            foo: NotRequired[str]
            bar: NotRequired[int]
            baz: NotRequired[float]
            qux: NotRequired[bool]

        schema = to_json_schema(MyTypedDict1)

        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "string"},
                "bar": {"type": "integer"},
                "baz": {"type": "number"},
                "qux": {"type": "boolean"},
            },
            "required": [],
        }

    def test_special(self):
        @dataclass
        class Schema:
            foo: None

        schema = to_json_schema(Schema)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "null"},
            },
            "required": ["foo"],
        }

        @dataclass
        class SchemaAny:
            foo: Any

        schema = to_json_schema(SchemaAny)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {},
            },
            "required": ["foo"],
        }

        @dataclass
        class UnionNone:
            foo: Union[None, None]

        schema = to_json_schema(UnionNone)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "null"},
            },
            "required": ["foo"],
        }

    def test_cache_hit(self):
        @dataclass
        class Schema:
            foo: str

        schema = to_json_schema(Schema)
        schema2 = to_json_schema(Schema)

        assert schema is schema2


# This function converts a dataclass to a TypedDict
def dataclass_to_typeddict(dc_cls: type):
    """Convert a dataclass to a TypedDict."""
    # Fetch type hints of the dataclass
    type_hints = get_type_hints(dc_cls)

    # Extract the fields and their types
    typeddict_fields: Dict = {
        field.name: type_hints[field.name] for field in fields(dc_cls)
    }

    # Create the TypedDict dynamically
    return TypedDict(f"{dc_cls.__name__}Dict", typeddict_fields)  # type: ignore



from __future__ import annotations

import logging
import os
from dataclasses import asdict, is_dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Generic, TypeVar, cast

import yaml

from .generate_yaml import dataclass_to_yaml
from .validate import to_json_schema, validate

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

# Needs the string escaping to work at runtime as _typeshed is not a real module
D = TypeVar("D", bound="DataclassInstance")

log = logging.getLogger(__name__)

__all__ = [
    "EYConf",
]


class EYConf(Generic[D]):
    """Configuration class.

    This class is used to generate a default configuration file from a schema
    represented by a dataclass.

    It allows to generate, validate and load a configuration file.
    """

    path: Path
    _schema: type[D]
    _data: D
    _json_schema: dict

    def __init__(
        self,
        schema: type[D],
        path: str | Path,
    ):

        if isinstance(path, str):
            path = Path(path)
        self.path = path

        # At the moment we only support dataclass classes
        if not is_dataclass(schema) or not isinstance(schema, type):
            raise ValueError(
                "Schema must be a dataclass class. Instances are not supported yet."
            )
        self._schema = schema

        # Generate default configuration if it does not exist
        if not self.path.exists():
            self.write_default()

        # Create schema
        self._json_schema = to_json_schema(self._schema)

        # Load the configuration file
        self._data = self._load()
        print(self._data)

    def write_default(self):
        """Generate default yaml configuration."""
        if self.path.exists():
            log.warning(f"Configuration file {self.path} already exists. Overwriting!")

        yaml_str = dataclass_to_yaml(self._schema)
        os.makedirs(self.path.parent, exist_ok=True)
        with open(self.path, "w") as f:
            f.write(yaml_str)
            f.write("\n")  # Add a newline at the end of the file
        log.info(f"Configuration file created at '{self.path.absolute()}'")

    def _load(self) -> D:
        """Load the configuration file and validate it against the schema."""
        log.info(f"Loading config file: {self.path.absolute()}")

        if not self.path.exists():
            raise FileNotFoundError(
                f"Configuration file '{self.path.absolute()}' not found. Please generate with `write_default()`."
            )

        # Load the config file
        with open(self.path, "r") as file:
            data = yaml.safe_load(file)

        # Will raise ConfigurationError if the data does not comply with the schema
        validate(data, self._json_schema)

        return cast(D, dataclass_from_dict(self._schema, data))

    def refresh(self):
        """Reload the configuration file."""
        self._data = self._load()

    def __getattr__(self, key: str):
        """Get an item from the configuration."""
        return getattr(self._data, key)

    def __str__(self):
        """Return a custom formatted string representation of the configuration data."""
        data_dict = asdict(self._data)
        class_name = type(self).__name__
        memory_address = hex(id(self))
        prefix = f"<{class_name} object at {memory_address} loaded from {self.path.absolute()}>:\n"
        s = "\n".join(
            "  " + line for line in self._pretty_format(data_dict).splitlines()
        )
        return f"{prefix}{s}"

    def _pretty_format(self, data, indent=0):
        """Format the dict with pretty indentation."""
        result = []
        for key, value in data.items():
            if isinstance(value, dict):
                result.append(" " * indent + f"{key}:")
                result.append(self._pretty_format(value, indent + 4))
            else:
                result.append(" " * indent + f"{key}: {value}")
        return "\n".join(result)

    __repr__ = __str__  # Use the same formatted string for repr, if desired


def dataclass_from_dict(klass, dikt):
    try:
        fieldtypes = klass.__annotations__
        return klass(**{f: dataclass_from_dict(fieldtypes[f], dikt[f]) for f in dikt})
    except AttributeError:
        if isinstance(dikt, (tuple, list)):
            return [dataclass_from_dict(klass.__args__[0], f) for f in dikt]
        return dikt



from dataclasses import dataclass
from pathlib import Path
import pytest
from eyconf import EYConf


@pytest.fixture
def tmp_config_path(tmp_path) -> Path:
    config_file_path = tmp_path / "config.yml"
    return config_file_path


def test_invalid_schema(tmp_config_path):

    with pytest.raises(ValueError):
        EYConf(int, tmp_config_path)  # type: ignore


def test_write_default(tmp_config_path):
    from eyconf import EYConf

    @dataclass
    class Config:
        int_field: int = 42
        str_field: str = "Hello, World!"

    conf = EYConf(Config, tmp_config_path)
    assert conf._schema == Config
    assert conf.path == tmp_config_path
    assert conf.path.exists()
    assert conf.path.read_text() == "int_field: 42\nstr_field: Hello, World!\n"


def test_load_existing(tmp_config_path):
    from eyconf import EYConf

    @dataclass
    class Config:
        int_field: int = 42
        str_field: str = "Hello, World!"

    with open(tmp_config_path, "w") as f:
        f.write("int_field: 20\nstr_field: Another value!\n")

    conf = EYConf(Config, tmp_config_path)

    assert conf.path == tmp_config_path
    assert conf.path.read_text() == "int_field: 20\nstr_field: Another value!\n"
    assert conf.int_field == 20
    assert conf.str_field == "Another value!"

    # Test reload
    with open(tmp_config_path, "w") as f:
        f.write("int_field: 30\nstr_field: Another value!\n")
    conf.refresh()
    assert conf.int_field == 30


def test_nested_getitem(tmp_config_path):
    from eyconf import EYConf

    @dataclass
    class Nested:
        int_field: int = 42
        str_field: str = "Hello, World!"

    @dataclass
    class Parent:
        nested: Nested
        other_field: str = "Hello, World!"

    conf = EYConf(Parent, tmp_config_path)

    assert type(conf.nested) == Nested



from dataclasses import dataclass, field
from typing import Annotated, Optional, Union

import logging
import yaml

from eyconf.generate_yaml import dataclass_to_yaml

log = logging.getLogger(__name__)


class TestGenerateDefault:
    """Test the generation of default yaml from a dataclass."""

    def test_basic(self):
        """Test the basic functionality."""

        @dataclass
        class Basic:
            int_field: int = 42
            str_field: str = "Hello, World!"
            bool_field: bool = True
            float_field: float = 3.14

        yaml_str = dataclass_to_yaml(Basic)
        assert (
            yaml_str
            == "int_field: 42\nstr_field: Hello, World!\nbool_field: true\nfloat_field: 3.14"
        )
        assert yaml.safe_load(yaml_str) == {
            "int_field": 42,
            "str_field": "Hello, World!",
            "bool_field": True,
            "float_field": 3.14,
        }

    def test_docstring(self):
        """Test the basic functionality."""

        @dataclass
        class SingleLineDocstring:
            """This is a docstring."""

            pass

        yaml_str = dataclass_to_yaml(SingleLineDocstring)
        assert yaml_str == "# This is a docstring.\n"
        assert yaml.safe_load(yaml_str) == None

        @dataclass
        class MultiLineDocstring:
            """This is a docstring that is longer than 80 characters and should be split into multiple lines."""

            pass

        yaml_str = dataclass_to_yaml(MultiLineDocstring)
        assert (
            yaml_str
            == "# This is a docstring that is longer than 80 characters and should be split into\n# multiple lines.\n"
        )
        assert yaml.safe_load(yaml_str) == None

        @dataclass
        class NoDocstring:
            pass

        yaml_str = dataclass_to_yaml(NoDocstring)
        assert yaml_str == ""
        assert yaml.safe_load(yaml_str) == None

    def test_annotation_docstring(self):
        """Test the basic functionality."""

        @dataclass
        class VariableDocstring:
            """This is a module docstring."""

            int_field: Annotated[int, "This is a docstring for the int field."] = 42

        yaml_str = dataclass_to_yaml(VariableDocstring)
        print(yaml_str)
        assert (
            yaml_str
            == "# This is a module docstring.\n\n# This is a docstring for the int field.\nint_field: 42"
        )

        @dataclass
        class MultiLineVariableDocstring:
            """This is a module docstring."""

            int_field: Annotated[int, "This is a docstring", "Test"] = 42

        yaml_str = dataclass_to_yaml(MultiLineVariableDocstring)
        print(yaml_str)
        assert (
            yaml_str
            == "# This is a module docstring.\n\n# This is a docstring\n# Test\nint_field: 42"
        )
        assert yaml.safe_load(yaml_str) == {"int_field": 42}

    def test_optional(self):

        @dataclass
        class Optionals:
            int_field: Optional[int]
            str_field: str | None
            other_field: Union[str, None]

        yaml_str = dataclass_to_yaml(Optionals)
        print(yaml_str)
        assert yaml_str == "int_field: null\nstr_field: null\nother_field: null"
        assert yaml.safe_load(yaml_str) == {
            "int_field": None,
            "str_field": None,
            "other_field": None,
        }

    def test_optional_with_default(self):

        @dataclass
        class OptionalsWithDefault:
            int_field: Optional[int] = 42
            str_field: str | None = "test"
            other_field: Union[str, None] = None

        yaml_str = dataclass_to_yaml(OptionalsWithDefault)
        assert yaml_str == "int_field: 42\nstr_field: test\nother_field: null"
        assert yaml.safe_load(yaml_str) == {
            "int_field": 42,
            "str_field": "test",
            "other_field": None,
        }

    def test_nested(self):
        """Nested dataclasses should raise an error."""

        @dataclass
        class Nested:
            int_field: int = 42
            str_field: str = "Hello, World!"

        @dataclass
        class Parent:
            nested: Nested
            other_field: str = "Hello, World!"

        yaml_str = dataclass_to_yaml(Parent)
        print(yaml_str)
        assert (
            yaml_str
            == "nested:\n  int_field: 42\n  str_field: Hello, World!\n\nother_field: Hello, World!"
        )
        assert yaml.safe_load(yaml_str) == {
            "nested": {"int_field": 42, "str_field": "Hello, World!"},
            "other_field": "Hello, World!",
        }

    def test_deep_nested(self):

        @dataclass
        class Nested:
            int_field: int = 42
            str_field: str = "Hello, World!"

        @dataclass
        class Parent:
            nested: Nested
            other_field: str = "Hello, World!"

        @dataclass
        class GrandParent:
            parent: Parent
            other_field: str = "Hello, World!"

        yaml_str = dataclass_to_yaml(GrandParent)
        print(yaml_str)
        assert (
            yaml_str
            == "parent:\n  nested:\n    int_field: 42\n    str_field: Hello, World!\n\n  other_field: Hello, World!\n\nother_field: Hello, World!"
        )
        assert yaml.safe_load(yaml_str) == {
            "parent": {
                "nested": {"int_field": 42, "str_field": "Hello, World!"},
                "other_field": "Hello, World!",
            },
            "other_field": "Hello, World!",
        }

    def test_list(self):
        """Test a list of dataclasses."""

        @dataclass
        class Lists:
            int_list: Optional[list[int]]
            str_list: list[str] = field(default_factory=lambda: ["a", "b", "c"])

        yaml_str = dataclass_to_yaml(Lists)
        print(yaml_str)
        assert yaml_str == "int_list: null\nstr_list:\n  - a\n  - b\n  - c"
        assert yaml.safe_load(yaml_str) == {
            "int_list": None,
            "str_list": ["a", "b", "c"],
        }

    def test_complex(self):
        """Test a more complex dataclass."""

        @dataclass
        class Movie:
            title: str = "Jurassic Park"
            year: int = 1993

        @dataclass
        class Director:
            name: str = "Steven Spielberg"
            movies: list[Movie] = field(
                default_factory=lambda: [
                    Movie(title="Jurassic Park", year=1993),
                    Movie(title="Indiana Jones", year=1981),
                ]
            )

        yaml_str = dataclass_to_yaml(Director)
        yaml_parsed = yaml.safe_load(yaml_str)

        assert yaml_parsed == {
            "name": "Steven Spielberg",
            "movies": [
                {"title": "Jurassic Park", "year": 1993},
                {"title": "Indiana Jones", "year": 1981},
            ],
        }

    def test_recursive(self):
        """Test a recursive dataclass."""

        @dataclass
        class Recursive:
            next: Optional["Recursive"] = None

        yaml_str = dataclass_to_yaml(Recursive)
        print(yaml_str)
        assert yaml_str == "next: null"
        assert yaml.safe_load(yaml_str) == {"next": None}



from collections.abc import Sequence as ABCSequence
from dataclasses import is_dataclass
from functools import lru_cache
from types import NoneType, UnionType
from typing import (
    Any,
    Dict,
    Literal,
    Sequence,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from jsonschema import Draft202012Validator
from typing_extensions import NotRequired

__all__ = ["to_json_schema", "primitives"]

primitives: dict[type[object], "str"] = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
}


@lru_cache(maxsize=None)
def to_json_schema(type: type, check_schema: bool = True) -> dict:
    """Convert a TypedDict or dataclass to a JSON schema.

    Parameters
    ----------
    type : type
        The TypedDict or dataclass to convert.
    check_schema : bool
        Whether to check the schema for validity.

    Raises
    ------
    ValueError
        If the type is not supported.
    SchemaError
        If the schema is invalid (should not happen).
    """
    schema = {
        "type": "object",
        "properties": {},
        "required": [],
    }

    # Get type hints for the TypedDict
    type_hints = get_type_hints(type, include_extras=True)

    # Add the type hints to the schema
    for field_name, field_type in type_hints.items():
        p, r = __convert_type_to_schema(field_type)
        schema["properties"][field_name] = p

        if r:
            schema["required"].append(field_name)

    if check_schema:
        Draft202012Validator.check_schema(schema)

    return schema


SchemaType = Union[
    dict[str, Any],
    dict[str, str],
    dict[Any, Any],
]


def __convert_type_to_schema(
    field_type: type,
) -> tuple[SchemaType, bool]:
    """
    Convert a type to a JSON schema.

    Parameters
    ----------
    field_type : type
        The type to convert.

    Returns
    -------
    schema : dict | None
        The JSON schema. None if the type is not supported or is is resolves to optional none value.
    is_required : bool
        Whether the field is required.
    """
    is_required = True

    origin = get_origin(field_type)

    # Handle Literal
    if origin is Literal:
        allowed_values = get_args(field_type)
        return {
            "type": __infer_type_from_values(allowed_values),
            "enum": list(allowed_values),
        }, is_required

    # Handle NotRequired
    if origin is NotRequired:
        field_type = get_args(field_type)[0]
        is_required = False
        return __convert_type_to_schema(field_type)[0], is_required

    # Handler union
    if origin is Union or origin is UnionType:
        allowed_types = set(get_args(field_type))
        if NoneType in allowed_types:
            is_required = False
            allowed_types.remove(NoneType)
        if None in allowed_types:
            is_required = False
            allowed_types.remove(None)
        if type(None) in allowed_types:
            is_required = False
            allowed_types.remove(type(None))

        if len(allowed_types) > 1:
            return {
                "anyOf": [
                    __convert_type_to_schema(allowed_type)[0]
                    for allowed_type in allowed_types
                ]
            }, is_required
        elif len(allowed_types) == 1:
            t, _ = __convert_type_to_schema(allowed_types.pop())
            return t, is_required
        else:
            raise ValueError("Union type must have at least one type!")

    # Handle sequence types
    if origin in [list, set, tuple, Sequence, ABCSequence]:
        return {
            "type": "array",
            "items": __convert_type_to_schema(get_args(field_type)[0])[0],
        }, is_required

    # Handle TypedDict and dataclasses
    if issubclass(field_type, Dict) or is_dataclass(field_type):
        return to_json_schema(field_type), is_required

    # Handle other types
    match = primitives.get(field_type)

    if match:
        return {"type": match}, is_required

    if field_type is None or field_type is NoneType:
        return {"type": "null"}, is_required

    if field_type is Any:
        return {}, is_required

    raise ValueError(f"Unsupported type: {field_type}")


def __infer_type_from_values(values: tuple | list):

    # Create set of types
    types = {type(value) for value in values}
    if len(types) > 1:
        raise ValueError("Literal values must all be of the same type")

    return primitives[types.pop()]
